import os, sys
import numpy as np
import scipy.interpolate
import scipy.optimize
import astropy.io.fits as fitsio
import astropy.wcs.utils
from astropy.wcs import WCS

from tools import *
from plotting import *

class WISP_Source():

    def __init__(self,par_num,obj_num,grism,data_dir,output_dir,
                 smooth=3,maglimit=24,limit=50,min_prior=0.):

        self.par_num    = par_num
        self.obj_num    = obj_num
        self.grism      = grism
        self.smooth     = smooth
        self.limit      = limit
        self.maglimit   = maglimit
        self.min_prior  = min_prior
        self.data_dir   = data_dir
        self.output_dir = output_dir
        self.par_dir    = os.path.join(self.data_dir,'Par%i/' % self.par_num)
        self.direct_dir = os.path.join(self.par_dir,'DATA/DIRECT_GRISM/')
        self.grism_dir  = os.path.join(self.par_dir,'%s_DRIZZLE/' % self.grism)

        self.check_input()
        self.set_output()

        self.set_direct()
        self.set_catalog()
        self.set_entry()
        self.check_files()

        self.get_response_curves()
        self.get_direct()
        self.get_grism()
        self.get_grism1D()
        self.pixfrac_corr = self.direct_pixfrac / self.grism_pixfrac

        self.set_direct_contams()
        self.set_grism_contams()
        self.get_profiles()
        self.get_prior()

    def check_input(self):

        if not os.path.exists(self.data_dir):
            raise Exception("No data_dir found at %s."%self.data_dir)

        if not os.path.exists(self.par_dir):
            raise Exception("No Par%i found at %s."%(self.par_num,self.par_dir))

        if self.grism not in ['G102','G141']:
            raise Exception("Incorrect grism. Please choose from 'G102' or 'G141'.")

        if self.smooth % 2 == 0:
            raise Exception('Provide an odd number of pixels to smooth over.')

    def set_output(self):

        os.system('mkdir -p %s/Par%i/plots'      % (self.output_dir,self.par_num))
        os.system('mkdir -p %s/Par%i/clean'      % (self.output_dir,self.par_num))
        os.system('mkdir -p %s/Par%i/stamps'     % (self.output_dir,self.par_num))
        os.system('mkdir -p %s/Par%i/contams'    % (self.output_dir,self.par_num))

    def set_direct(self):

        if self.grism == 'G102':
            if   os.path.isfile(self.par_dir+'DATA/DIRECT_GRISM/F110W_sci.fits'): self.direct = 'F110'
            else: raise Exception('No F110W direct image found for %s'%self.grism)

        if self.grism == 'G141':
            if   os.path.isfile(self.par_dir+'DATA/DIRECT_GRISM/F140W_sci.fits'): self.direct = 'F140'
            elif os.path.isfile(self.par_dir+'DATA/DIRECT_GRISM/F160W_sci.fits'): self.direct = 'F160'
            else: raise Exception('No F140W or F160W direct image found for %s'%self.grism)

    def set_catalog(self):

        self._catalog = np.genfromtxt('%s/fin_%s.cat' % (self.direct_dir,self.direct),
                                        dtype=[('RA_DEC_NAME','>S25'),('OBJ_NUM',int),('X_IMAGE',float),('Y_IMAGE',float),
                                               ('A_IMAGE',float),('B_IMAGE',float),('THETA_IMAGE',float),('X_WORLD',float),
                                               ('Y_WORLD',float),('A_WORLD',float),('B_WORLD',float),('THETA_WORLD',float),
                                               ('MAG',float),('MAGERR_AUTO',float),('CLASS_STAR',float),('FLAGS',int)])
        bbox_x, bbox_y = bboxEllipse(self._catalog['A_IMAGE'],self._catalog['B_IMAGE'],self._catalog['THETA_IMAGE'])

        self.catalog = np.recarray(len(self._catalog),
                                    dtype=[('PAR_NUM',int),('DIRECT','|S4'),('GRISM','|S4')] +
                                           self._catalog.dtype.descr +
                                           [('BBOX_X',float),('BBOX_Y',float)])
        for x in self._catalog.dtype.names: self.catalog[x] = self._catalog[x]

        self.catalog['PAR_NUM'] = self.par_num
        self.catalog['DIRECT']  = self.direct
        self.catalog['GRISM']   = self.grism
        self.catalog['BBOX_X']  = 2*bbox_x
        self.catalog['BBOX_Y']  = 2*bbox_y

    def set_entry(self):

        entry = self.catalog[self.catalog['OBJ_NUM'] == self.obj_num]
        if len(entry) > 0:
            self.entry = entry[0]
            self.obj_num = self.entry['OBJ_NUM']
        else:
            raise Exception("No object with ID#%i found in Par%i." % (obj_num,par_num))

    def check_files(self):

        if not os.path.isfile('%s/%sW_sci.fits' % (self.direct_dir,self.direct)):
            raise Exception('No %s direct image found at %s.'%(self.direct,self.direct_dir))

        if not os.path.isfile('%s/aXeWFC3_%s_mef_ID%i.fits' % (self.grism_dir,self.grism,self.obj_num)):
            raise Exception('No OBJ#%s grism data found at %s.'%(self.obj_num,self.grism_dir))

    def get_response_curves(self,fill_value=1e-20):

        pivot_l = {'F110':11534.00, 'F140':13922.979, 'F160':15369.161}
        waves, sens = np.genfromtxt('config/HST_WFC3_IR.%sW.dat'%self.direct,unpack=True)
        self.direct_pivot = pivot_l[self.direct]
        sens[sens==0] = fill_value
        self.direct_response = scipy.interpolate.interp1d(waves,sens,bounds_error=False,fill_value=fill_value)

        sens_file = fitsio.getdata('config/WFC3.%s.1st.sens.fits' % self.grism)
        waves, sens = sens_file['WAVELENGTH'], sens_file['SENSITIVITY']
        sens[sens==0] = fill_value
        self.grism_response = scipy.interpolate.interp1d(waves,sens,bounds_error=False,fill_value=fill_value)

    def get_direct(self):

        self.direct_hdu = fitsio.open('%s/%sW_sci.fits' % (self.direct_dir,self.direct))
        self.direct_hdr = self.direct_hdu[0].header
        self.direct_img = self.direct_hdu[0].data
        self.direct_wcs = WCS(self.direct_hdr)
        self.direct_pixfrac = self.direct_hdr['D001SCAL']

        masked = mask_direct_image(self.direct_img, self.direct_hdr, self.catalog, fill_value=-99.)
        self.direct_bkgd = np.median(masked[masked != -99.])
        self.direct_img_nobkgd = self.direct_img.copy() - self.direct_bkgd

    def get_grism(self):

        self.grism_hdu = fitsio.open('%s/aXeWFC3_%s_mef_ID%i.fits' % (self.grism_dir,self.grism,self.obj_num))
        self.grism_hdr = self.grism_hdu['SCI'].header
        self.grism_img = self.grism_hdu['SCI'].data
        self.grism_err = self.grism_hdu['ERR'].data
        self.grism_len = self.grism_hdr['NAXIS1']
        self.grism_pixfrac = self.grism_hdr['CDELT2']

        self.grism_mask = self.grism_err==0
        self.grism_img  = np.ma.masked_array(self.grism_img,mask=self.grism_mask,fill_value=np.NaN)
        self.grism_err  = np.ma.masked_array(self.grism_err,mask=self.grism_mask,fill_value=np.NaN)

        self.grism_model = np.zeros_like(self.grism_img)
        self.grism_img_s = np.zeros_like(self.grism_img)
        self.grism_err_s = np.zeros_like(self.grism_img)

        n = (self.smooth-1)/2
        self.grism_img_s[:,n:-n] =            moving_average(self.grism_img,   s=self.smooth)
        self.grism_err_s[:,n:-n] = np.ma.sqrt(moving_average(self.grism_err**2,s=self.smooth) / self.smooth)

    def get_grism1D(self):

        self.grism1D_img, self.grism1D_wht = np.ma.average(self.grism_img, weights=np.ma.power(self.grism_err,-2), axis=-1, returned=True)
        self.grism1D_err  = np.ma.power(self.grism1D_wht,-0.5)
        self.grism1D_mask = self.grism1D_img.mask

        self.x  = np.arange(len(self.grism1D_img))
        grism1D_img_interpfn = scipy.interpolate.interp1d(self.x[~self.grism1D_mask],self.grism1D_img[~self.grism1D_mask],bounds_error=False,fill_value=0)
        grism1D_err_interpfn = scipy.interpolate.interp1d(self.x[~self.grism1D_mask],self.grism1D_err[~self.grism1D_mask],bounds_error=False,fill_value=0)

        self.xf = np.linspace(min(self.x),max(self.x),len(self.x)*100.)
        self.grism1D_img = grism1D_img_interpfn(self.xf)
        self.grism1D_err = grism1D_err_interpfn(self.xf)
        self.grism1D_img = np.ma.masked_array(self.grism1D_img,mask=self.grism1D_err==0,fill_value=np.NaN)
        self.grism1D_err = np.ma.masked_array(self.grism1D_err,mask=self.grism1D_err==0,fill_value=np.NaN)

    def set_direct_contams(self):

        dx, dy = 5*self.entry['BBOX_X'], 5*self.entry['BBOX_Y']
        self.direct_contams = get_direct_contams(self.catalog,self.entry,dx,dy)

    def set_grism_contams(self):

        dy, dx = np.asarray(self.grism_img.shape)
        dx = dx / self.pixfrac_corr
        dy = dy / self.pixfrac_corr
        self.grism_contams = get_grism_contams(self.catalog,self.entry,dx,dy/2.)
        cond = (self.grism_contams['MAG'] <= self.maglimit) & (self.grism_contams['OBJ_NUM']<1000)
        self.grism_contams = self.grism_contams[cond]
        if len(self.grism_contams) > self.limit:
            raise Exception('%i contaminants exceeds the maximum limit (%s).'%(len(self.grism_contams),self.limit))

    def get_profiles(self):

        mk_stamps(self.entry,self)
        for contam in self.grism_contams: mk_stamps(contam,self)

        self.d_prof =  mk_stamp_profile(self.entry,self.pixfrac_corr,self.output_dir,self.direct)
        self.c_prof = [mk_stamp_profile(contam,self.pixfrac_corr,self.output_dir,self.direct) for contam in self.grism_contams]
        self.d_cen  = self.grism_img.shape[0] / 2.
        self.c_cen  = (self.grism_contams['Y_IMAGE'] - self.entry['Y_IMAGE']) * self.pixfrac_corr + self.d_cen
        self.model_args = (self.d_prof, self.d_cen, self.c_prof, self.c_cen)

    def get_prior(self):

        waves = get_waves(np.arange(self.grism_len)+1, self.grism_hdr)
        self.prior = np.zeros((len(self.grism_contams)+1,len(waves)))
        self.prior[0,:] = self.grism_response(waves)
        for i,contam in enumerate(self.grism_contams):
            del_x = (contam['X_IMAGE'] - self.entry['X_IMAGE']) * self.pixfrac_corr
            waves = get_waves(np.arange(self.grism_len) - del_x, self.grism_hdr)
            self.prior[i+1,:] = self.grism_response(waves)
        self.prior = self.prior / np.max(self.prior[0,:])
        self.prior = np.clip(self.prior+self.min_prior, 0, 1)

        self.prior1D = np.sum(self.prior,axis=0)
        self.prior1D = self.prior1D / self.prior1D[0]

    def minimize1D(self,pars,x,ydata,yerr,prior):

        model    = profile_model(x,pars,self.model_args,prior)
        invsig2  = np.ma.power(yerr,-2)
        do_sum   = (ydata-model)**2 * invsig2 - np.ma.log(2*np.pi*invsig2)
        lnl      = 0.5 * np.ma.sum(do_sum.filled(1e5))
        return lnl

    def get_subpx_shifts(self):

        x0 = [0,] + [0.01,0] * (len(self.grism_contams)+1)
        bounds = [(-5,5),] + [(0,5),(-2,2)] * (len(self.grism_contams)+1)
        self.subpx_pars = scipy.optimize.minimize(self.minimize1D,x0=x0,bounds=bounds,args=(self.xf,self.grism1D_img,self.grism1D_err,self.prior1D))['x']
        print "Sub-pixelling (%s) done for Par#%i Obj#%i" % (self.grism,self.entry['PAR_NUM'],self.entry['OBJ_NUM'])

    def minimize(self,pars_,x,ydata,yerr,prior):

        pars = self.subpx_pars.copy()
        pars[1::2] = pars_
        return self.minimize1D(pars,x,ydata,yerr,prior)

    def clean_grism(self):

        x0 = self.subpx_pars[1::2]
        bounds = [(0,5),] * len(x0)
        for i in range(self.grism_len):
            ydata, yerr = self.grism_img_s[:,i], self.grism_err_s[:,i]
            if not all(yerr==0.) and any(np.isfinite(ydata)):
                pars_ = scipy.optimize.minimize(self.minimize,x0=x0,bounds=bounds,args=(self.x,ydata,yerr,self.prior[:,i]))['x']
                pars = self.subpx_pars.copy()
                pars[1::2] = pars_
                self.grism_model[:,i] = contam_model(self.x,pars,self.c_prof,self.c_cen,self.prior[:,i])
        print "Grism (%s) cleaning done for Par#%i Obj#%i" % (self.grism,self.entry['PAR_NUM'],self.entry['OBJ_NUM'])
        self.grism_clean = self.grism_img - self.grism_model

    def get_spec1D(self):

        norm = scipy.integrate.quad(self.d_prof,0,self.grism_img.shape[0],epsabs=1e-6,epsrel=1e-6,limit=250)[0]
        x = np.arange(self.grism_img.shape[0])
        d_prof = direct_model(x,self.subpx_pars,self.d_prof,self.d_cen)
        d_prof = d_prof / norm

        self.spec_1D = np.ma.zeros(self.grism_len,fill_value=np.NaN)
        for i in range(self.grism_len):
            self.spec_1D[i] = np.ma.sum(self.grism_img[:,i] * d_prof) - self.subpx_pars[0]

        self.waves = get_waves(np.arange(self.grism_len)+1, self.grism_hdr)
        sens = self.grism_response(self.waves)
        cond = (sens!=0)
        self.flux_1D = np.zeros(self.grism_len)
        self.flux_1D[cond] = self.spec_1D[cond] / sens[cond]
        self.flux_1D[~cond] = 0.

    def save_output(self):

        # Save the contaminants (SEx catalog)
        data = np.vstack((self.grism_contams['OBJ_NUM'],self.grism_contams['X_IMAGE'],self.grism_contams['Y_IMAGE'],self.grism_contams['A_IMAGE'],self.grism_contams['B_IMAGE'],self.grism_contams['THETA_IMAGE'],self.grism_contams['X_WORLD'],self.grism_contams['Y_WORLD'],self.grism_contams['MAG'],self.grism_contams['MAGERR_AUTO'],self.grism_contams['CLASS_STAR'])).T
        hdr = "%5s%10s%10s%10s%10s%10s%15s%15s%10s%10s%10s" % ('OBJ_NUM','X_IMAGE','Y_IMAGE','A_IMAGE','B_IMAGE','THETA_IMAGE','X_WORLD','Y_WORLD','MAG','MAGERR_AUTO','CLASS_STAR')
        fmt = "%5i%10.4f%10.4f%10.4f%10.4f%10.4f%15.6e%15.6f%10.4f%10.4f%10.2f"
        np.savetxt(self.output_dir+'Par%i/contams/contams_%s_ID%i.fits' % (self.par_num,self.grism,self.obj_num),data,header=hdr,fmt=fmt)

        # Save the continuum and subpixel offsets
        with open(self.output_dir+'Par%i/contams/offsets_%s_ID%i.fits' % (self.par_num,self.grism,self.obj_num),'w') as f:
            f.write('Continuum: %15.6e\n' % self.subpx_pars[0])
            f.write('Object: %15.6e %15.6e\n' % (self.subpx_pars[1],self.subpx_pars[2]))
            for i,(x,y) in enumerate(zip(self.subpx_pars[3::2],self.subpx_pars[4::2])):
                f.write('Contam#%i: %15.6e %15.6e\n' % (i+1,x,y))

        # Make and save profile plot
        self.profile_fig = plot_profiles(self)
        self.profile_fig.savefig(self.output_dir+'Par%i/plots/plot_profile_%s_ID%i.png' % (self.par_num,self.grism,self.obj_num))

        # Make and save the grism plot
        self.grism_fig = plot_grisms(self)
        self.grism_fig.savefig(self.output_dir+'Par%i/plots/plot_grism_%s_ID%i.png' % (self.par_num,self.grism,self.obj_num))

        # Make and save the grism FITS output
        hdulist = fitsio.HDUList([])
        for data,extname in zip([self.grism_img,self.grism_err,self.grism_model,self.grism_clean],['SCI','ERR','MODEL','CLEAN']):
            data = data.filled(fill_value=np.NaN)
            hdu = fitsio.PrimaryHDU(header=self.grism_hdr,data=data)
            hdu.name = extname
            hdulist.append(hdu)
        hdulist.writeto(self.output_dir+'Par%i/clean/aXeWFC3_%s_clean_ID%i.fits' % (self.par_num,self.grism,self.obj_num),overwrite=True)

    def process(self):

        self.get_subpx_shifts()
        self.clean_grism()
        self.get_spec1D()
        self.save_output()

if __name__ == '__main__':
    
    print "No main() defined."
